@@ -21,7 +21,7 @@
  * OUT OF OR IN CONNECTION WITH THE SOFTWARE OR THE USE OR OTHER DEALINGS IN THE
  * SOFTWARE.
  */
-#define BOOST_TEST_MODULE ordered_map_tests
+#define BOOST_TEST_MODULE robin_map_tests
 #define BOOST_TEST_DYN_LINK
 
 
