@@ -115,23 +115,25 @@ BOOST_AUTO_TEST_CASE_TEMPLATE(test_insert, HMap, test_types) {
 }
 
 BOOST_AUTO_TEST_CASE(test_range_insert) {
+    // create a vector<std::pair> of values to insert, insert part of them in the map, check values
     const int nb_values = 1000;
-    std::vector<std::pair<int, int>> values;
+    std::vector<std::pair<int, int>> values_to_insert(nb_values);
     for(int i = 0; i < nb_values; i++) {
-        values.push_back(std::make_pair(i, i+1));
+        values_to_insert[i] = std::make_pair(i, i+1);
     }
     
     
     tsl::robin_map<int, int> map = {{-1, 1}, {-2, 2}};
-    map.insert(std::next(values.begin(), 10), values.end() - 5);
+    map.insert(std::next(values_to_insert.begin(), 10), values_to_insert.end() - 5);
+    
     
     BOOST_CHECK_EQUAL(map.size(), 987);
     
-    BOOST_CHECK_EQUAL(map[-1], 1);
-    BOOST_CHECK_EQUAL(map[-2], 2);
+    BOOST_CHECK_EQUAL(map.at(-1), 1);
+    BOOST_CHECK_EQUAL(map.at(-2), 2);
     
     for(int i = 10; i < nb_values - 5; i++) {
-        BOOST_CHECK_EQUAL(map[i], i+1);
+        BOOST_CHECK_EQUAL(map.at(i), i+1);
     }
 }
 
@@ -160,6 +162,38 @@ BOOST_AUTO_TEST_CASE(test_insert_with_hint) {
     BOOST_CHECK_EQUAL(map.size(), 5);
 }
 
+/**
+ * emplace_hint
+ */
+BOOST_AUTO_TEST_CASE(test_emplace_hint) {
+    tsl::robin_map<int, int> map{{1, 0}, {2, 1}, {3, 2}};
+    
+    // Wrong hint
+    BOOST_CHECK(map.emplace_hint(map.find(2), std::piecewise_construct, 
+                                              std::forward_as_tuple(3), std::forward_as_tuple(4)) == map.find(3));
+    
+    // Good hint
+    BOOST_CHECK(map.emplace_hint(map.find(2), std::piecewise_construct, 
+                                              std::forward_as_tuple(2), std::forward_as_tuple(4)) == map.find(2));
+    
+    // end() hint
+    BOOST_CHECK(map.emplace_hint(map.find(10), std::piecewise_construct, 
+                                               std::forward_as_tuple(2), std::forward_as_tuple(4)) == map.find(2));
+    
+    BOOST_CHECK_EQUAL(map.size(), 3);
+    
+    
+    // end() hint, new value
+    BOOST_CHECK_EQUAL(map.emplace_hint(map.find(10), std::piecewise_construct, 
+                                                    std::forward_as_tuple(4), std::forward_as_tuple(3))->first, 4);
+    
+    // Wrong hint, new value
+    BOOST_CHECK_EQUAL(map.emplace_hint(map.find(2), std::piecewise_construct, 
+                                                    std::forward_as_tuple(5), std::forward_as_tuple(4))->first, 5);
+    
+    BOOST_CHECK_EQUAL(map.size(), 5);
+}
+
 /**
  * emplace
  */
@@ -170,16 +204,16 @@ BOOST_AUTO_TEST_CASE(test_emplace) {
     
     
     std::tie(it, inserted) = map.emplace(std::piecewise_construct,
-                                            std::forward_as_tuple(10),
-                                            std::forward_as_tuple(1));
+                                         std::forward_as_tuple(10),
+                                         std::forward_as_tuple(1));
     BOOST_CHECK_EQUAL(it->first, 10);
     BOOST_CHECK_EQUAL(it->second, move_only_test(1));
     BOOST_CHECK(inserted);
     
     
     std::tie(it, inserted) = map.emplace(std::piecewise_construct,
-                                            std::forward_as_tuple(10),
-                                            std::forward_as_tuple(3));
+                                         std::forward_as_tuple(10),
+                                         std::forward_as_tuple(3));
     BOOST_CHECK_EQUAL(it->first, 10);
     BOOST_CHECK_EQUAL(it->second, move_only_test(1));
     BOOST_CHECK(!inserted);
@@ -208,6 +242,7 @@ BOOST_AUTO_TEST_CASE(test_try_emplace) {
 }
 
 BOOST_AUTO_TEST_CASE(test_try_emplace_2) {
+    // Insert x values with try_emplace, insert them again, check with find.
     tsl::robin_map<std::string, move_only_test> map;
     tsl::robin_map<std::string, move_only_test>::iterator it;
     bool inserted;
@@ -305,7 +340,7 @@ BOOST_AUTO_TEST_CASE(test_insert_or_assign_hint) {
  * erase
  */
 BOOST_AUTO_TEST_CASE(test_range_erase_all) {
-    // insert x values, delete all
+    // insert x values, delete all with iterators
     using HMap = tsl::robin_map<std::string, std::int64_t>;
     
     const std::size_t nb_values = 1000;
@@ -317,7 +352,7 @@ BOOST_AUTO_TEST_CASE(test_range_erase_all) {
 }
 
 BOOST_AUTO_TEST_CASE(test_range_erase) {
-    // insert x values, delete all except 10 first and 780 last values
+    // insert x values, delete all with iterators except 10 first and 780 last values
     using HMap = tsl::robin_map<std::string, std::int64_t>;
     
     const std::size_t nb_values = 1000;
@@ -337,7 +372,7 @@ BOOST_AUTO_TEST_CASE(test_range_erase) {
 }
 
 BOOST_AUTO_TEST_CASE_TEMPLATE(test_erase_loop, HMap, test_types) {
-    // insert x values, delete all one by one
+    // insert x values, delete all one by one with iterator
     std::size_t nb_values = 1000;
     
     HMap map = utils::get_filled_hash_map<HMap>(nb_values);
@@ -359,18 +394,18 @@ BOOST_AUTO_TEST_CASE_TEMPLATE(test_erase_loop, HMap, test_types) {
 }
 
 BOOST_AUTO_TEST_CASE_TEMPLATE(test_erase_loop_range, HMap, test_types) {
-    // insert x values, delete all five by five
-    const std::size_t range = 5;
+    // insert x values, delete all five by five with iterators
+    const std::size_t hop = 5;
     std::size_t nb_values = 1000;
     
-    BOOST_REQUIRE_EQUAL(nb_values % range, 0);
+    BOOST_REQUIRE_EQUAL(nb_values % hop, 0);
     
     HMap map = utils::get_filled_hash_map<HMap>(nb_values);
     
     auto it = map.begin();
     while(it != map.end()) {
-        it = map.erase(it, std::next(it, range));
-        nb_values -= range;
+        it = map.erase(it, std::next(it, hop));
+        nb_values -= hop;
         
         BOOST_CHECK_EQUAL(map.size(), nb_values);
     }
@@ -383,7 +418,7 @@ BOOST_AUTO_TEST_CASE_TEMPLATE(test_insert_erase_insert, HMap, test_types) {
     using key_t = typename HMap::key_type; using value_t = typename HMap:: mapped_type;
     
     const std::size_t nb_values = 2000;
-    HMap map;
+    HMap map(10);
     typename HMap::iterator it;
     bool inserted;
     
@@ -436,6 +471,7 @@ BOOST_AUTO_TEST_CASE_TEMPLATE(test_insert_erase_insert, HMap, test_types) {
 }
 
 BOOST_AUTO_TEST_CASE(test_range_erase_same_iterators) {
+    // insert x values, test erase with same iterator as each parameter, check if returned mutable iterator is valid.
     const std::size_t nb_values = 100;
     auto map = utils::get_filled_hash_map<tsl::robin_map<std::int64_t, std::int64_t>>(nb_values);
     
@@ -451,10 +487,75 @@ BOOST_AUTO_TEST_CASE(test_range_erase_same_iterators) {
     BOOST_CHECK_EQUAL(it_const.value(), -100);
 }
 
+/**
+ * max_load_factor
+ */
+BOOST_AUTO_TEST_CASE(test_max_load_factor_extreme_factors) {
+    tsl::robin_map<std::int64_t, std::int64_t> map;
+    
+    map.max_load_factor(0.0f);
+    BOOST_CHECK_GT(map.max_load_factor(), 0.0f);
+    
+    map.max_load_factor(10.0f);
+    BOOST_CHECK_LT(map.max_load_factor(), 1.0f);
+}
+
+/**
+ * min_load_factor
+ */
+BOOST_AUTO_TEST_CASE(test_min_load_factor_extreme_factors) {
+    tsl::robin_map<std::int64_t, std::int64_t> map;
+    
+    BOOST_CHECK_EQUAL(map.min_load_factor(), 0.0f);
+    BOOST_CHECK_LT(map.min_load_factor(), map.max_load_factor());
+    
+    map.min_load_factor(-10.0f);
+    BOOST_CHECK_EQUAL(map.min_load_factor(), 0.0f);
+    
+    map.min_load_factor(0.9f);
+    map.max_load_factor(0.1f);
+    
+    // max_load_factor should always be > min_load_factor. 
+    // Factors should have been clamped.
+    BOOST_CHECK_LT(map.min_load_factor(), map.max_load_factor());
+}
+
+BOOST_AUTO_TEST_CASE(test_min_load_factor) {
+    tsl::robin_map<std::int64_t, std::int64_t> map;
+    
+    map.min_load_factor(0.15f);
+    BOOST_CHECK_EQUAL(map.min_load_factor(), 0.15f);
+    
+    map.max_load_factor(0.5f);
+    BOOST_CHECK_EQUAL(map.max_load_factor(), 0.5f);
+    
+    
+    map.rehash(100);
+    for(std::size_t i = 0; i < map.bucket_count()/2; i++) {
+        map.insert({utils::get_key<std::int64_t>(i), 
+                    utils::get_value<std::int64_t>(i)});
+    }
+    
+    BOOST_CHECK_GT(map.load_factor(), map.min_load_factor());
+    // Can't use BOOST_CHECK_CLOSE with -fno-exceptions
+    BOOST_CHECK(std::abs(map.load_factor() - 0.5f) < 0.005f);
+    
+    
+    while(map.load_factor() >= map.min_load_factor()) {
+        map.erase(map.begin());
+    }
+    
+    // Shrink is done on insert.
+    map.insert({utils::get_key<std::int64_t>(map.bucket_count()), 
+                utils::get_value<std::int64_t>(map.bucket_count())});
+    BOOST_CHECK_GT(map.load_factor(), map.min_load_factor());
+}
+
 /**
  * rehash
  */
 BOOST_AUTO_TEST_CASE(test_rehash_empty) {
+    // test rehash(0), test find/erase/insert on map.
     const std::size_t nb_values = 100;
     auto map = utils::get_filled_hash_map<tsl::robin_map<std::int64_t, std::int64_t>>(nb_values);
     
@@ -463,49 +564,63 @@ BOOST_AUTO_TEST_CASE(test_rehash_empty) {
     
     map.clear();
     BOOST_CHECK_EQUAL(map.bucket_count(), bucket_count);
+    BOOST_CHECK(map.empty());
     
     map.rehash(0);
     BOOST_CHECK_EQUAL(map.bucket_count(), 0);
+    BOOST_CHECK(map.empty());
+    
+    
+    BOOST_CHECK(map.find(1) == map.end());
+    BOOST_CHECK_EQUAL(map.erase(1), 0);
+    BOOST_CHECK(map.insert({1, 10}).second);
+    BOOST_CHECK_EQUAL(map.at(1), 10);
 }
 
 
 /**
  * operator== and operator!=
  */
-BOOST_AUTO_TEST_CASE_TEMPLATE(test_compare, HMap, test_types) {
-    // create 3 maps, 2 are the same, compare maps
-    using key_t = typename HMap::key_type; using value_t = typename HMap:: mapped_type;
+BOOST_AUTO_TEST_CASE(test_compare) {
+    const tsl::robin_map<std::string, std::int64_t> map1 = {{"a", 1}, {"e", 5}, {"d", 4}, {"c", 3}, {"b", 2}};
+    const tsl::robin_map<std::string, std::int64_t> map1_copy = {{"e", 5}, {"c", 3}, {"b", 2}, {"a", 1}, {"d", 4}};
+    const tsl::robin_map<std::string, std::int64_t> map2 = {{"e", 5}, {"c", 3}, {"b", 2}, {"a", 1}, {"d", 4}, {"f", 6}};
+    const tsl::robin_map<std::string, std::int64_t> map3 = {{"e", 5}, {"c", 3}, {"b", 2}, {"a", 1}};
+    const tsl::robin_map<std::string, std::int64_t> map4 = {{"a", 1}, {"e", 5}, {"d", 4}, {"c", 3}, {"b", 26}};
+    const tsl::robin_map<std::string, std::int64_t> map5 = {{"a", 1}, {"e", 5}, {"d", 4}, {"c", 3}, {"z", 2}};
     
-    const std::size_t nb_values = 1000;
-    HMap map_1_1;
-    HMap map_1_2;
-    HMap map_2_1;
+    BOOST_CHECK(map1 == map1_copy);
+    BOOST_CHECK(map1_copy == map1);
     
-    for(std::size_t i = 0; i < nb_values; i++) {
-        map_1_1.insert({utils::get_key<key_t>(i), utils::get_value<value_t>(i)});
-        if(i != 0) {
-            map_2_1.insert({utils::get_key<key_t>(i), utils::get_value<value_t>(i)});
-        }
-    }
+    BOOST_CHECK(map1 != map2);
+    BOOST_CHECK(map2 != map1);
     
-    // Same as map_1_1 but insertion order inverted
-    for(std::size_t i = nb_values; i != 0; i--) {
-        map_1_2.insert({utils::get_key<key_t>(i-1), utils::get_value<value_t>(i-1)});
-    }
+    BOOST_CHECK(map1 != map3);
+    BOOST_CHECK(map3 != map1);
     
+    BOOST_CHECK(map1 != map4);
+    BOOST_CHECK(map4 != map1);
     
-    BOOST_CHECK_EQUAL(map_1_1.size(), nb_values);
-    BOOST_CHECK_EQUAL(map_1_2.size(), nb_values);
-    BOOST_CHECK_EQUAL(map_2_1.size(), nb_values-1);
+    BOOST_CHECK(map1 != map5);
+    BOOST_CHECK(map5 != map1);
     
-    BOOST_CHECK(map_1_1 == map_1_2);
-    BOOST_CHECK(map_1_2 == map_1_1);
+    BOOST_CHECK(map2 != map3);
+    BOOST_CHECK(map3 != map2);
     
-    BOOST_CHECK(map_1_1 != map_2_1);
-    BOOST_CHECK(map_2_1 != map_1_1);
+    BOOST_CHECK(map2 != map4);
+    BOOST_CHECK(map4 != map2);
     
-    BOOST_CHECK(map_1_2 != map_2_1);
-    BOOST_CHECK(map_2_1 != map_1_2);
+    BOOST_CHECK(map2 != map5);
+    BOOST_CHECK(map5 != map2);
+    
+    BOOST_CHECK(map3 != map4);
+    BOOST_CHECK(map4 != map3);
+    
+    BOOST_CHECK(map3 != map5);
+    BOOST_CHECK(map5 != map3);
+    
+    BOOST_CHECK(map4 != map5);
+    BOOST_CHECK(map5 != map4);
 }
 
 
@@ -515,7 +630,7 @@ BOOST_AUTO_TEST_CASE_TEMPLATE(test_compare, HMap, test_types) {
  * clear
  */
 BOOST_AUTO_TEST_CASE(test_clear) {
-    // insert x values, clear map
+    // insert x values, clear map, test insert
     using HMap = tsl::robin_map<std::int64_t, std::int64_t>;
     
     const std::size_t nb_values = 1000;
@@ -535,8 +650,8 @@ BOOST_AUTO_TEST_CASE(test_clear) {
 /**
  * iterator.value()
  */
-BOOST_AUTO_TEST_CASE(test_modify_value) {
-    // insert x values, modify value of even keys, check values
+BOOST_AUTO_TEST_CASE(test_modify_value_through_iterator) {
+    // insert x values, modify value of even keys with iterators, check values
     const std::size_t nb_values = 100;
     auto map = utils::get_filled_hash_map<tsl::robin_map<std::int64_t, std::int64_t>>(nb_values);
     
@@ -560,36 +675,43 @@ BOOST_AUTO_TEST_CASE(test_modify_value) {
  * constructor
  */
 BOOST_AUTO_TEST_CASE(test_extreme_bucket_count_value_construction) {
-    BOOST_CHECK_THROW((tsl::robin_map<int, int, std::hash<int>, std::equal_to<int>, 
+    TSL_RH_CHECK_THROW((tsl::robin_map<int, int, std::hash<int>, std::equal_to<int>, 
                                       std::allocator<std::pair<int, int>>, false,
                                       tsl::rh::power_of_two_growth_policy<2>>
                             (std::numeric_limits<std::size_t>::max())), std::length_error);
     
-    BOOST_CHECK_THROW((tsl::robin_map<int, int, std::hash<int>, std::equal_to<int>, 
+    TSL_RH_CHECK_THROW((tsl::robin_map<int, int, std::hash<int>, std::equal_to<int>, 
                                       std::allocator<std::pair<int, int>>, false, 
                                       tsl::rh::power_of_two_growth_policy<2>>
                             (std::numeric_limits<std::size_t>::max()/2 + 1)), std::length_error);
     
     
     
-    BOOST_CHECK_THROW((tsl::robin_map<int, int, std::hash<int>, std::equal_to<int>, 
+    TSL_RH_CHECK_THROW((tsl::robin_map<int, int, std::hash<int>, std::equal_to<int>, 
                                       std::allocator<std::pair<int, int>>, false, 
                                       tsl::rh::prime_growth_policy>
                             (std::numeric_limits<std::size_t>::max())), std::length_error);
     
-    BOOST_CHECK_THROW((tsl::robin_map<int, int, std::hash<int>, std::equal_to<int>, 
+    TSL_RH_CHECK_THROW((tsl::robin_map<int, int, std::hash<int>, std::equal_to<int>, 
                                       std::allocator<std::pair<int, int>>, false, 
                                       tsl::rh::prime_growth_policy>
                             (std::numeric_limits<std::size_t>::max()/2)), std::length_error);
     
     
     
-    BOOST_CHECK_THROW((tsl::robin_map<int, int, std::hash<int>, std::equal_to<int>, 
+    TSL_RH_CHECK_THROW((tsl::robin_map<int, int, std::hash<int>, std::equal_to<int>, 
                                       std::allocator<std::pair<int, int>>, false, 
                                       tsl::rh::mod_growth_policy<>>
                             (std::numeric_limits<std::size_t>::max())), std::length_error);
 }
 
+BOOST_AUTO_TEST_CASE(test_range_construct) {
+    tsl::robin_map<int, int> map = {{2, 1}, {1, 0}, {3, 2}};
+    
+    tsl::robin_map<int, int> map2(map.begin(), map.end());
+    tsl::robin_map<int, int> map3(map.cbegin(), map.cend());
+}
+
 
 /**
  * operator=(std::initializer_list)
@@ -603,6 +725,9 @@ BOOST_AUTO_TEST_CASE(test_assign_operator) {
     BOOST_CHECK_EQUAL(map.at(1), 3);
     BOOST_CHECK_EQUAL(map.at(2), 4);
     BOOST_CHECK(map.find(0) == map.end());
+    
+    map = {};
+    BOOST_CHECK(map.empty());
 }
 
 
@@ -610,7 +735,7 @@ BOOST_AUTO_TEST_CASE(test_assign_operator) {
  * move/copy constructor/operator
  */
 BOOST_AUTO_TEST_CASE(test_move_constructor) {
-    // insert x values in map, move map into map_move, check map and map_move, 
+    // insert x values in map, move map into map_move with move constructor, check map and map_move, 
     // insert additional values in map_move, check map_move
     using HMap = tsl::robin_map<std::string, move_only_test>;
     
@@ -631,9 +756,19 @@ BOOST_AUTO_TEST_CASE(test_move_constructor) {
     BOOST_CHECK(map_move == utils::get_filled_hash_map<HMap>(nb_values*2));
 }
 
+BOOST_AUTO_TEST_CASE(test_move_constructor_empty) {
+    tsl::robin_map<std::string, move_only_test> map(0);
+    tsl::robin_map<std::string, move_only_test> map_move(std::move(map));
+    
+    BOOST_CHECK(map.empty());
+    BOOST_CHECK(map_move.empty());
+    
+    BOOST_CHECK(map.find("") == map.end());
+    BOOST_CHECK(map_move.find("") == map_move.end());
+}
 
 BOOST_AUTO_TEST_CASE(test_move_operator) {
-    // insert x values in map, move map into map_move, check map and map_move, 
+    // insert x values in map, move map into map_move with move operator, check map and map_move, 
     // insert additional values in map_move, check map_move
     using HMap = tsl::robin_map<std::string, move_only_test>;
     
@@ -655,6 +790,18 @@ BOOST_AUTO_TEST_CASE(test_move_operator) {
     BOOST_CHECK(map_move == utils::get_filled_hash_map<HMap>(nb_values*2));
 }
 
+BOOST_AUTO_TEST_CASE(test_move_operator_empty) {
+    tsl::robin_map<std::string, move_only_test> map(0);
+    tsl::robin_map<std::string, move_only_test> map_move;
+    map_move = (std::move(map));
+    
+    BOOST_CHECK(map.empty());
+    BOOST_CHECK(map_move.empty());
+    
+    BOOST_CHECK(map.find("") == map.end());
+    BOOST_CHECK(map_move.find("") == map_move.end());
+}
+
 BOOST_AUTO_TEST_CASE(test_reassign_moved_object_move_constructor) {
     using HMap = tsl::robin_map<std::string, std::string>;
     
@@ -681,25 +828,6 @@ BOOST_AUTO_TEST_CASE(test_reassign_moved_object_move_operator) {
     BOOST_CHECK(map == (HMap({{"Key4", "Value4"}, {"Key5", "Value5"}})));
 }
 
-BOOST_AUTO_TEST_CASE(test_copy_constructor_operator) {
-    using HMap = tsl::robin_map<std::string, std::string, mod_hash<9>>;
-    
-    
-    const std::size_t nb_values = 100;
-    HMap map = utils::get_filled_hash_map<HMap>(nb_values);
-    
-    HMap map_copy = map;
-    HMap map_copy2(map);
-    HMap map_copy3 = utils::get_filled_hash_map<HMap>(1);
-    map_copy3 = map;
-    
-    BOOST_CHECK(map == map_copy);
-    map.clear();
-    
-    BOOST_CHECK(map_copy == map_copy2);
-    BOOST_CHECK(map_copy == map_copy3);
-}
-
 BOOST_AUTO_TEST_CASE(test_use_after_move_constructor) {
     using HMap = tsl::robin_map<std::string, move_only_test>;
     
@@ -745,24 +873,77 @@ BOOST_AUTO_TEST_CASE(test_use_after_move_operator) {
     BOOST_CHECK(map == map_move);
 }
 
+BOOST_AUTO_TEST_CASE(test_copy_constructor_and_operator) {
+    using HMap = tsl::robin_map<std::string, std::string, mod_hash<9>>;
+    
+    
+    const std::size_t nb_values = 100;
+    HMap map = utils::get_filled_hash_map<HMap>(nb_values);
+    
+    HMap map_copy = map;
+    HMap map_copy2(map);
+    HMap map_copy3 = utils::get_filled_hash_map<HMap>(1);
+    map_copy3 = map;
+    
+    BOOST_CHECK(map == map_copy);
+    map.clear();
+    
+    BOOST_CHECK(map_copy == map_copy2);
+    BOOST_CHECK(map_copy == map_copy3);
+}
+
+BOOST_AUTO_TEST_CASE(test_copy_constructor_empty) {
+    tsl::robin_map<std::string, int> map(0);
+    tsl::robin_map<std::string, int> map_copy(map);
+    
+    BOOST_CHECK(map.empty());
+    BOOST_CHECK(map_copy.empty());
+    
+    BOOST_CHECK(map.find("") == map.end());
+    BOOST_CHECK(map_copy.find("") == map_copy.end());
+}
+
+BOOST_AUTO_TEST_CASE(test_copy_operator_empty) {
+    tsl::robin_map<std::string, int> map(0);
+    tsl::robin_map<std::string, int> map_copy(16);
+    map_copy = map;
+    
+    BOOST_CHECK(map.empty());
+    BOOST_CHECK(map_copy.empty());
+    
+    BOOST_CHECK(map.find("") == map.end());
+    BOOST_CHECK(map_copy.find("") == map_copy.end());
+}
+
 
 /**
  * at
  */
 BOOST_AUTO_TEST_CASE(test_at) {
     // insert x values, use at for known and unknown values.
-    tsl::robin_map<std::int64_t, std::int64_t> map = {{0, 10}, {-2, 20}};
+    const tsl::robin_map<std::int64_t, std::int64_t> map = {{0, 10}, {-2, 20}};
     
     BOOST_CHECK_EQUAL(map.at(0), 10);
     BOOST_CHECK_EQUAL(map.at(-2), 20);
-    BOOST_CHECK_THROW(map.at(1), std::out_of_range);
+    TSL_RH_CHECK_THROW(map.at(1), std::out_of_range);
+}
+
+/**
+ * contains
+ */
+BOOST_AUTO_TEST_CASE(test_contains) {
+    tsl::robin_map<std::int64_t, std::int64_t> map = {{0, 10}, {-2, 20}};
+    
+    BOOST_CHECK(map.contains(0));
+    BOOST_CHECK(map.contains(-2));
+    BOOST_CHECK(!map.contains(-3));
 }
 
 /**
  * equal_range
  */
 BOOST_AUTO_TEST_CASE(test_equal_range) {
-    tsl::robin_map<std::int64_t, std::int64_t> map = {{0, 10}, {-2, 20}};
+    const tsl::robin_map<std::int64_t, std::int64_t> map = {{0, 10}, {-2, 20}};
     
     auto it_pair = map.equal_range(0);
     BOOST_REQUIRE_EQUAL(std::distance(it_pair.first, it_pair.second), 1);
@@ -810,6 +991,63 @@ BOOST_AUTO_TEST_CASE(test_swap) {
     BOOST_CHECK(map2 == (tsl::robin_map<std::int64_t, std::int64_t>{{1, 10}, {8, 80}, {3, 30}, {4, 40}}));
 }
 
+BOOST_AUTO_TEST_CASE(test_swap_empty) {
+    tsl::robin_map<std::int64_t, std::int64_t> map = {{1, 10}, {8, 80}, {3, 30}};
+    tsl::robin_map<std::int64_t, std::int64_t> map2;
+    
+    using std::swap;
+    swap(map, map2);
+    
+    BOOST_CHECK(map == (tsl::robin_map<std::int64_t, std::int64_t>{}));
+    BOOST_CHECK(map2 == (tsl::robin_map<std::int64_t, std::int64_t>{{1, 10}, {8, 80}, {3, 30}}));
+    
+    map.insert({6, 60});
+    map2.insert({4, 40});
+    
+    BOOST_CHECK(map == (tsl::robin_map<std::int64_t, std::int64_t>{{6, 60}}));
+    BOOST_CHECK(map2 == (tsl::robin_map<std::int64_t, std::int64_t>{{1, 10}, {8, 80}, {3, 30}, {4, 40}}));
+}
+
+/**
+ * KeyEqual
+ */
+BOOST_AUTO_TEST_CASE(test_key_equal) {
+    // Use a KeyEqual and Hash where any odd unsigned number 'x' is equal to 'x-1'.
+    // Make sure that KeyEqual is called (and not ==).
+    struct hash {
+        std::size_t operator()(std::uint64_t v) const {
+            if(v%2u == 1u) {
+                return std::hash<std::uint64_t>()(v-1);
+            }
+            else {
+                return std::hash<std::uint64_t>()(v);
+            }
+        }
+    };
+    
+    struct key_equal {
+        bool operator()(std::uint64_t lhs, std::uint64_t rhs) const {
+            if(lhs%2u == 1u) {
+                lhs--;
+            }
+            
+            if(rhs%2u == 1u) {
+                rhs--;
+            }
+            
+            return lhs == rhs;
+        }
+    };
+    
+    tsl::robin_map<std::uint64_t, std::uint64_t, hash, key_equal> map;
+    BOOST_CHECK(map.insert({2, 10}).second);
+    BOOST_CHECK_EQUAL(map.at(2), 10);
+    BOOST_CHECK_EQUAL(map.at(3), 10);
+    BOOST_CHECK(!map.insert({3, 10}).second);
+    
+    BOOST_CHECK_EQUAL(map.size(), 1);
+}
+
 
 
 /**
@@ -873,7 +1111,7 @@ BOOST_AUTO_TEST_CASE(test_heterogeneous_lookups) {
     
     BOOST_CHECK_EQUAL(map.at(addr1), 4);
     BOOST_CHECK_EQUAL(map.at(addr2), 5);
-    BOOST_CHECK_THROW(map.at(addr_unknown), std::out_of_range);
+    TSL_RH_CHECK_THROW(map.at(addr_unknown), std::out_of_range);
     
     
     BOOST_REQUIRE(map.find(addr1) != map.end());
@@ -908,6 +1146,7 @@ BOOST_AUTO_TEST_CASE(test_empty_map) {
     
     BOOST_CHECK_EQUAL(map.bucket_count(), 0);
     BOOST_CHECK_EQUAL(map.size(), 0);
+    BOOST_CHECK_EQUAL(map.load_factor(), 0);
     BOOST_CHECK(map.empty());
     
     BOOST_CHECK(map.begin() == map.end());
@@ -920,8 +1159,11 @@ BOOST_AUTO_TEST_CASE(test_empty_map) {
     BOOST_CHECK_EQUAL(map.count(""), 0);
     BOOST_CHECK_EQUAL(map.count("test"), 0);
     
-    BOOST_CHECK_THROW(map.at(""), std::out_of_range);
-    BOOST_CHECK_THROW(map.at("test"), std::out_of_range);
+    BOOST_CHECK(!map.contains(""));
+    BOOST_CHECK(!map.contains("test"));
+    
+    TSL_RH_CHECK_THROW(map.at(""), std::out_of_range);
+    TSL_RH_CHECK_THROW(map.at("test"), std::out_of_range);
     
     auto range = map.equal_range("test");
     BOOST_CHECK(range.first == range.second);
@@ -958,7 +1200,16 @@ BOOST_AUTO_TEST_CASE(test_precalculated_hash) {
     BOOST_CHECK_EQUAL(map_const.at(3, map_const.hash_function()(3)), -3);
     
     BOOST_REQUIRE_NE(map.hash_function()(2), map.hash_function()(3));
-    BOOST_CHECK_THROW(map.at(3, map.hash_function()(2)), std::out_of_range);
+    TSL_RH_CHECK_THROW(map.at(3, map.hash_function()(2)), std::out_of_range);
+    
+    /**
+     * contains
+     */
+    BOOST_CHECK(map.contains(3, map.hash_function()(3)));
+    BOOST_CHECK(map_const.contains(3, map_const.hash_function()(3)));
+    
+    BOOST_REQUIRE_NE(map.hash_function()(2), map.hash_function()(3));
+    BOOST_CHECK(!map.contains(3, map.hash_function()(2)));
     
     /**
      * count
